"""File transfer package model."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass


@dataclass
class FileTransferPackage:
    """
    Data from a file binary request response.

    :ivar previous_hash: Hash of the previous chunk
    :vartype previous_hash: bytes
    :ivar data: Requested chunk
    :vartype data: bytes
    :ivar current_hash: Hash of the current chunk
    :vartype current_hash: bytes
    """

    previous_hash: bytes
    data: bytes
    current_hash: bytes
